{# highlight_bad_id.html #}

{# DEFUNCT FILE #}

{% extends "base.html" %}

{% block content %}

    {# include "highlight_nav.html" #}

    <h1>Invalid highlight ID</h1>
    <p>Query {{ highlight_id }} is invalid. Please create or choose another highlight.</p>

{% endblock %}
