{# email_researcher_approval.html #}
{% extends 'base_email.html' %}

{% comment %}
    template parameters:
        consent_mode: ConsentMode
        contact_request: ContactRequest
        patient_lookup: PatientLookup
        study: Study
{% endcomment %}

{% block content %}

    <p>Dear {{ study.lead_researcher.profile.get_salutation }},</p>

    <p>Good news! Following your contact request, you’ve been approved to
    contact a patient about your study.</p>

    <p><b>Please read the attached approval letter CAREFULLY.</b></p>

    <p>Patient:</p>

    {% include "patient_table.html" with patient_lookup=patient_lookup consent_mode=consent_mode include_sex=True %}

    <p>Study:</p>
    {% include 'study_table.html' with study=study contact_request=contact_request %}

    <p>Yours sincerely,</p>

    <p><i>The CPFT Research Database Computer</i></p>

{% endblock %}
