{# email_rdbm_good_progress.html #}
{% extends 'base_email.html' %}

{% comment %}
    template parameters:
        id: int
        response: str
        explanation: str
{% endcomment %}

{% block content %}

    <p>A clinician agreed to a request from the CPFT Research Database
    consent-mode computer. The contact request ID was {{ id }}.
    Their response was <b>{{ response }} – {{ explanation }}</b>.
    Charity money needs sending.

{% endblock %}
